public float[][][] getData(String file) {
  /* getData() loads coordinates from a file 
      and returns a coordinate array */
  
  // If file doesn't exist, return pre-existing mesh
  if (file == null)
    return mesh;
  
  // Based on the # of lines, determine size of array
  String lines[] = loadStrings(file);
  int numSteps = int(sqrt(lines.length-1)+1);
  float mesh[][][] = new float[numSteps+1][numSteps][3];
  distFromCenter = new float[numSteps][numSteps];
  
  String substrings[] = new String[3];
  float sphericalCoords[] = new float[3];

  for (int i=0; i<lines.length; i++) {
    // Convert comma-delimited ASCII data into floats
    substrings = split(lines[i],",");
    sphericalCoords = float(substrings);
    
    // Convert spherical coordinates to Cartesian
    if (int(i/numSteps)%2 == 0 || plotLinear) {
      mesh[int(i/numSteps)][i%numSteps] = toCartesian(sphericalCoords);
      distFromCenter[int(i/numSteps)][i%numSteps] = sphericalCoords[2];
    }
    else {
      mesh[int(i/numSteps)][numSteps-i%numSteps-1] = toCartesian(sphericalCoords);
      distFromCenter[int(i/numSteps)][numSteps-i%numSteps-1] = sphericalCoords[2];
    }
  }

  // Copy the unmodified mesh over to another array
  originalMesh = new float[numSteps+1][numSteps][3];
  for (int i=0; i<numSteps; i++) {
    for (int j=0; j<numSteps-1; j++) {
      for (int k=0; k<3; k++) {
        originalMesh[i][j][k] = mesh[i][j][k];
      }
    }
  }

  for (int j=0; j<numSteps; j++) {
    // Connect the beginning and the end
    originalMesh[numSteps][j] = mesh[0][numSteps-1-j];
    mesh[numSteps][j] = mesh[0][numSteps-1-j];
  }

  for (int i=0; i<smoothingIterations; i++)
    mesh = smoothMesh(mesh);
  
  return mesh;
}

public float[] toCartesian(float input[]) {
  /* toCartesian[] converts a polar coordinate 
      to a cartesian coordinate in a float array*/
  float theta = radians(input[0]-90);
  float phi = radians(input[1]-90);
  float r = input[2];
  
  float cartCoords[] = new float[3];
  
  if (r>maximumR) {
    for (int i=0; i<3; i++) {
      cartCoords[i] = 0;
      return cartCoords;
    }
  }
    
  cartCoords[0] = r*cos(theta)*sin(phi);
  cartCoords[2] = r*sin(theta)*sin(phi);
  cartCoords[1] = r*cos(phi);
  
  return cartCoords;
}

public void drawProperties() {
  // Set drawing properties based on settings
  scale(zoomScale);
  
  // Draw axes
  if (showAxes) {
    strokeWeight(AXIS_THICKNESS);
    stroke(X_AXIS_COLOR);
    line(0,0,0,AXIS_LENGTH,0,0);
    stroke(Y_AXIS_COLOR);
    line(0,0,0,0,AXIS_LENGTH,0);
    stroke(Z_AXIS_COLOR);
    line(0,0,0,0,0,AXIS_LENGTH);
  }
  
  strokeWeight(LINE_THICKNESS);
  
  // Determine whether or not to fill
  if (wireframe)
    noFill();
  else
    fill(fillColor);
}

public boolean[][] determineFaces(float dists[][]) {
  /* Checks to see if any point is farther away from
      the center than the threshold value; if so, reject
      that point and don't plot it. */
      
  // Returns a boolean array; true means draw face, false means don't.
  boolean[][] returnBool = new boolean[dists.length-1][dists.length-1];
  
  // Initialize all entries to true
  for (int i=0; i < dists.length-1; i++) {
    for (int j=0; j < dists.length-1; j++) {
      returnBool[i][j] = true;
    }
  }
  
  for (int i=0; i < dists.length-1; i++) {
    for (int j=0; j < dists.length-1; j++) {
      if (dists[i][j] >= maximumR) {
        returnBool[i][j] = false;
        if (i != 0 && j != 0)
          returnBool[i-1][j-1] = false;
        if (i != 0)
          returnBool[i-1][j] = false;
        if (j != 0)
          returnBool[i][j-1] = false;
      }
    }
  }
  return returnBool;
}

public float distBetween(float point1[], float point2[]) {
  /* Finds the distance between two points defined by a 
      float array of length 3 (x, y, z)*/
  float sumOfSquares = 0;
  for (int i=0; i<3; i++) {
    sumOfSquares += pow(point1[i] - point2[i], 2);
  }
  return sqrt(sumOfSquares);
}

public float[][][] smoothMesh(float mesh[][][]) {
  /* smoothMesh() takes every point in the mesh and
      smooths it based on the 8 points directly surrounding it,
      with more weight on closer points. */
  float[][][] returnMesh = mesh;
  
  for (int i=0; i < mesh.length-2; i++) {
    for (int j=0; j < mesh[i].length-1; j++) {
      for (int k=0; k < 3; k++) {
        returnMesh[i][j][k] = mesh[i][j][k]/MAIN_WEIGHT;
        
        /* If i is 0, the points to the left are actually the points on
            the far right; looping around. */
        if (i == 0) {
          /* If j is 0 or the maximum value, there aren't 8 points to check, 
              so check the points that can be checked, with more weight*/
          if (j == 0) {
            returnMesh[i][j][k] += (mesh[mesh.length-2][mesh[i].length-j-2][k]+mesh[i+1][j][k]+mesh[mesh.length-2][mesh[i].length-j-1][k]+mesh[i+1][j+1][k])/DIAG_WEIGHT;
            returnMesh[i][j][k] += (mesh[i][j+1][k]+mesh[mesh[i].length-2][mesh[i].length-j-1][k]+mesh[i+1][j][k])/SIDE_WEIGHT2;
          }
          else if (j == mesh[i].length-1) {
            returnMesh[i][j][k] += (mesh[mesh.length-2][mesh[i].length-j-2][k]+mesh[i+1][j-1][k]+mesh[mesh.length-2][mesh[i].length-j][k]+mesh[i+1][j][k])/DIAG_WEIGHT;
            returnMesh[i][j][k] += (mesh[i][j-1][k]+mesh[i][j][k]+mesh[mesh[i].length-2][mesh[i].length-j-1][k]+mesh[i+1][j][k])/SIDE_WEIGHT;
          }
          /* If not at the boundary of the mesh, do this */
          else {
            returnMesh[i][j][k] += (mesh[mesh.length-2][mesh[i].length-j-2][k]+mesh[i+1][j-1][k]+mesh[mesh.length-2][mesh[i].length-j][k]+mesh[i+1][j+1][k])/DIAG_WEIGHT;
            returnMesh[i][j][k] += (mesh[i][j-1][k]+mesh[i][j+1][k]+mesh[mesh[i].length-2][mesh[i].length-j-1][k]+mesh[i+1][j][k])/SIDE_WEIGHT;
          }
        }
        else {
          /* If j is 0 or the maximum value, there aren't 8 points to check, 
              so check the points that can be checked, with more weight*/
          if (j == 0) {
            returnMesh[i][j][k] += (mesh[i-1][j][k]+mesh[i+1][j][k]+mesh[i-1][j+1][k]+mesh[i+1][j+1][k])/DIAG_WEIGHT;
            returnMesh[i][j][k] += (mesh[i][j+1][k]+mesh[i-1][j][k]+mesh[i+1][j][k])/SIDE_WEIGHT2;
          }
          else if (j == mesh[i].length-1) {
            returnMesh[i][j][k] += (mesh[i-1][j-1][k]+mesh[i+1][j-1][k]+mesh[i-1][j][k]+mesh[i+1][j][k])/DIAG_WEIGHT;
            returnMesh[i][j][k] += (mesh[i][j-1][k]+mesh[i][j][k]+mesh[i-1][j][k]+mesh[i+1][j][k])/SIDE_WEIGHT;
          }
          /* If not at the boundary of the mesh, do this */
          else {
            returnMesh[i][j][k] += (mesh[i-1][j-1][k]+mesh[i+1][j-1][k]+mesh[i-1][j+1][k]+mesh[i+1][j+1][k])/DIAG_WEIGHT;
            returnMesh[i][j][k] += (mesh[i][j-1][k]+mesh[i][j+1][k]+mesh[i-1][j][k]+mesh[i+1][j][k])/SIDE_WEIGHT;
          }
        }
      }
    }
  }
  
  for (int j=0; j<mesh[0].length-1; j++) {
    // Connect the beginning and the end
    returnMesh[mesh[0].length-1][j] = returnMesh[0][mesh[0].length-1-j];
  }
  
  return returnMesh;
}

void smoothMultiple(int times) {
  for (int i=0; i<mesh.length; i++) {
    for (int j=0; j<mesh[i].length; j++) {
      for (int k=0; k<3; k++) {
        mesh[i][j][k] = originalMesh[i][j][k];
      }
    }
  }
  
  for (int i=0; i<times; i++) {
    mesh = smoothMesh(mesh);
  }
}

public void drawMesh(float mesh[][][]) {
  /* drawMesh() plots the points and lines/surfaces
      based on graphics settings */
  
  // Determine whether or not to draw lines, based on settings
  if (showLines || wireframe || whichShape == SHAPES[0] || whichShape == SHAPES[1])
    stroke(lineColor);
  else
    noStroke();
  
  // Begin drawing
  for (int i = 0; i < meshBool.length; i++) {
    beginShape(whichShape);
    if (useTexture && !wireframe) {  // If textures enabled
      texture(img);
    
      for (int j = 0; j < meshBool.length; j++) {
        if (meshBool[i][j]) { // determined from determineFaces()
          u = (float)i/mesh.length;
          v = (float)j/mesh[i].length;
          normal(mesh[i][j][0], mesh[i][j][1], mesh[i][j][2]);
          vertex(mesh[i][j][0], mesh[i][j][1], mesh[i][j][2],u,v);
          
          u = (float)(i+1)/mesh.length;
          normal(mesh[i+1][j][0], mesh[i+1][j][1], mesh[i+1][j][2]);
          vertex(mesh[i+1][j][0], mesh[i+1][j][1], mesh[i+1][j][2],u,v);
        } else {  
          /* If meshBool is false for this coordinate, 
              don't draw a shape, start a new one */
          endShape();
          beginShape(whichShape);
          texture(img);
        }
      }
      endShape(); // End shape for this column
    }
    else { // If textures disabled, do the following
      for (int j = 0; j < meshBool.length; j++) {
        if (meshBool[i][j]) { // determined from determineFaces()
          vertex(mesh[i][j][0], mesh[i][j][1], mesh[i][j][2]);
          vertex(mesh[i+1][j][0], mesh[i+1][j][1], mesh[i+1][j][2]);
        } else {
          /* If meshBool is false for this coordinate, 
              don't draw a shape, start a new one */
          endShape();
          beginShape(whichShape);
        }
      }
      endShape(); // End shape for this column
    }
  }
}
